/**
 * Send "Hello" over USB every second.
 * 
 * Note: Only interrupt-driven UART API works with Zephyr CDC AMC,
 * not compatible with UART Async API (rip).
 * 
 * Helpful articles
 * https://devzone.nordicsemi.com/f/nordic-q-a/94627/uart-interrupt-tx
 * https://academy.nordicsemi.com/courses/nrf-connect-sdk-intermediate/lessons/lesson-3-adding-custom-board-support/topic/exercise-1/
 * https://docs.nordicsemi.com/bundle/ncs-2.4.3/page/zephyr/services/usb/uds_cdc_acm.html
 */

#include <zephyr/kernel.h>
#include <zephyr/drivers/uart.h>
#include <zephyr/usb/usb_device.h>
#include <zephyr/sys/ring_buffer.h>

//const struct device *uart_dev = DEVICE_DT_GET(DT_NODELABEL(uart0));
const struct device *uart_dev = DEVICE_DT_GET(DT_CHOSEN(zephyr_console));

#define RING_BUF_SIZE 1024
uint8_t ring_buffer[RING_BUF_SIZE];

struct ring_buf ringbuf;

const struct uart_config uart_cfg =
{
    .baudrate = 115200,
    .parity = UART_CFG_PARITY_NONE,
    .stop_bits = UART_CFG_STOP_BITS_1,
    .data_bits = UART_CFG_DATA_BITS_8,
    .flow_ctrl = UART_CFG_FLOW_CTRL_DTR_DSR
};

static void interrupt_handler(const struct device *dev, void *user_data)
{
	ARG_UNUSED(user_data);

	// Call update before processing a pending request
	// if interrupt request still in queue
	while (uart_irq_update(dev) && uart_irq_is_pending(dev))
	{
		if (uart_irq_tx_ready(dev))
		{
			uint8_t buffer[64];
			int rb_len, send_len;

			rb_len = ring_buf_get(&ringbuf, buffer, sizeof(buffer));

			// if nothing to send, disable tx interrupt
			if (!rb_len)
			{
				//LOG_DBG("Ring buffer empty, disable TX IRQ");
				uart_irq_tx_disable(dev);
				continue;
			}
			// else send ring buffer contents over UART
			send_len = uart_fifo_fill(dev, buffer, rb_len);

			// if not all of the bytes were sent, some were dropped.
			/*if (send_len < rb_len)
			{
				LOG_ERR("Drop %d bytes", rb_len - send_len);
			}*/

			//LOG_DBG("ringbuf -> tty fifo %d bytes", send_len);
		}
	}
}

void usb_enable_cb(enum usb_dc_status_code status, const uint8_t *param) {
    switch (status) {
        case USB_DC_CONFIGURED:
            printk("USB device configured\n");
            break;
        case USB_DC_CONNECTED:
            printk("USB device halted\n");
            break;
        case USB_DC_DISCONNECTED:
            printk("USB device disconnected\n");
            break;
        case USB_DC_ERROR:
            printk("USB error\n");
            break;
        case USB_DC_SET_HALT:
            printk("USB set halt\n");
            break;
        case USB_DC_CLEAR_HALT:
            printk("USB clear halt\n");
            break;
        default:
            break;
    }
}

int main(void)
{
    int retVal; 

    // is UART ready?
    if (!device_is_ready(uart_dev))
    {
        return 0;
    }

    // set UART configurations
    retVal = uart_configure(uart_dev, &uart_cfg);

	if (retVal == -ENOSYS)
    {
		printk("UART config options not supported.");
	}

    // enable the USB subsystem and associated hardware

    retVal = usb_enable(usb_enable_cb);

	if (retVal != 0)
    {
		printk("Failed to enable USB");
		return 0;
	}

    // sleep until UART data terminal is ready
    while (true)
	{
		uint32_t dtr = 0U; // DTR = Data Terminal Ready = is it ready to go?

		uart_line_ctrl_get(uart_dev, UART_LINE_CTRL_DTR, &dtr);

		if (dtr)
		{
			break;
		}
		else
		{
			// give CPU resources to low priority threads
			k_sleep(K_MSEC(100));
		}
	}

    // wait 100ms for the host to do all settings
	k_msleep(100);

	uart_irq_callback_set(uart_dev, interrupt_handler);

	// Here's the data to be written
	char awesome_data[8] = "Hello!\n\0";

	while(1)
	{
		// Delay for 1 second
		k_msleep(1000);

		// Add data to ring buffer
		ring_buf_put(&ringbuf, awesome_data, sizeof(awesome_data));

		// Enable the TX interrupt
		uart_irq_tx_enable(uart_dev);
	}

    return 0;
}
